@@ -1,6 +1,6 @@
 /* Copyright 2007 by Kim Minh Kaplan
  *
- * greyfix.c version 0.3.1
+ * greyfix.c version 0.3.2
  *
  * Postfix policy daemon designed to prevent spam using the
  * greylisting method.
@@ -13,6 +13,7 @@
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include <assert.h>
 #include <errno.h>
 #include <signal.h>
 #include <stdio.h>
@@ -38,7 +39,7 @@
  * This determines how many seconds of life are given to a record that
  * is created from a new mail [ip,from,to] triplet.  Note that the
  * window created by this setting for passing mails is reduced by the
- * amount set for $delay_mail_secs.  NOTE: See Also:
+ * amount set for DELAY_MAIL_SECS.  NOTE: See Also:
  * update_record_life and update_record_life_secs.
  */
 #define AUTO_RECORD_LIFE_SECS (5 * 3600) /* 5 hours */
@@ -94,6 +95,14 @@ static DBT dbdata = { 0 };
 static struct triplet_data triplet_data;
 
 static int debug_me = 0;
+static unsigned long greylist_delay = DELAY_MAIL_SECS;
+static unsigned long bloc_max_idle = AUTO_RECORD_LIFE_SECS;
+static unsigned long pass_max_idle = UPDATE_RECORD_LIFE_SECS;
+/* As we store IP addresses in Postfix's format, to obtain the network
+ address we first strip `ipv4_network_strip_bytes' numbers (between 0
+ and 4) then we apply `ipv4_network_mask' on the last byte. */
+static unsigned int ipv4_network_strip_bytes;
+static unsigned int ipv4_network_mask;
 
 /**********************************************************************
  * Berkeley DB routines
@@ -254,28 +263,74 @@ xrealloc(void *ptr, size_t size)
 static void
 build_triplet_key(const char *ip, const char *from, const char *to)
 {
+    int ipv4 = 1;
     const char *endip = strchr(ip, '\n'),
 	*endfrom = strchr(from, '\n'),
 	*endto = strchr(to, '\n');
-    size_t lenip = endip - ip,
-	lenfrom = endfrom - from,
+    size_t lenfrom = endfrom - from,
 	lento = endto - to;
-    size_t total = lenip + lenfrom + lento + 2;
+    size_t lenip, total;
     char *buf;
+    /* Mangle the IP address so that only the required prefix is used */
+    if (ipv4_network_strip_bytes > 0) {
+	const char *p = endip;
+	unsigned int i = ipv4_network_strip_bytes;
+	while (i && --p > ip)
+	    if (*p == '.')
+		i--;
+	if (i <= 1)
+	    endip = p;
+	else {
+	    ipv4 = 0;
+	    syslog(LOG_ERR, "Could not apply network strip");
+	}
+    }
+    lenip = endip - ip,
+    total = lenip + lenfrom + lento + 2;
     if (dbkey.ulen < total) {
 	dbkey.data = xrealloc(dbkey.data, total);
 	dbkey.ulen = total;
 	dbkey.flags = DB_DBT_USERMEM;
     }
-    dbkey.size = total;
     buf = (char*)dbkey.data;
-    memcpy(buf, ip, lenip);
-    buf += lenip;
+    if (ipv4 && ipv4_network_mask != 0xffU) {
+	/* Mask the last octet of the IP address */
+	char *q;
+	unsigned long byte;
+	const char *p = endip;
+	while (--p > ip)
+	    if (*p == '.')
+		break;
+	if (*p == '.')
+	    p++;
+	byte = strtoul(p, &q, 10);
+	if (p != q && q <= endip && byte < 256U) {
+	    size_t n = p - ip;
+	    memcpy(buf, ip, n);
+	    buf += n;
+	    /* XXX the byte we are subsituting can only be smaller
+	       than the original so no additional memory is needed. */
+	    n = sprintf(buf, "%u", byte & ipv4_network_mask);
+	    buf += n;
+	    assert(buf - (char *)dbkey.data <= lenip);
+	}
+	else
+	    ipv4 = 0;
+    }
+    if (! ipv4 || ipv4_network_mask == 0xffU) {
+	memcpy(buf, ip, lenip);
+	buf += lenip;
+    }
     *buf++ = 0;
+    if (debug_me)
+	syslog(LOG_DEBUG, "triplet effective IP: %s", dbkey.data);
     memcpy(buf, from, lenfrom);
     buf += lenfrom;
     *buf++ = 0;
     memcpy(buf, to, lento);
+    buf += lento;
+    dbkey.size = buf - (char *)dbkey.data;
+    assert(dbkey.size <= dbkey.ulen);
 }
 
 static void
@@ -451,8 +506,8 @@ static int
 triplet_expired_p(const struct triplet_data *data)
 {
     unsigned int age = triplet_data.access_time - triplet_data.create_time;
-    return triplet_data.pass_count == 0 && age > AUTO_RECORD_LIFE_SECS
-	|| triplet_data.pass_count != 0 && age > UPDATE_RECORD_LIFE_SECS;
+    return triplet_data.pass_count == 0 && age > bloc_max_idle
+	|| triplet_data.pass_count != 0 && age > pass_max_idle;
 }
 
 static void
@@ -469,8 +524,8 @@ process_smtp_rcpt()
     }
     delay = triplet_data.access_time - triplet_data.create_time;
     /* Block inbound mail that is from a previously unknown (ip, from, to) triplet */
-    if (delay < DELAY_MAIL_SECS) {
-	delay = DELAY_MAIL_SECS - delay;
+    if (delay < greylist_delay) {
+	delay = greylist_delay - delay;
 	triplet_data.block_count++;
 	printf("action=DEFER_IF_PERMIT Greylisted by " PACKAGE_STRING ", "
 	       "try again in %lu second%s."
@@ -495,11 +550,20 @@ signal_handler(int signal)
     exit(-1);
 }
 
+static int
+optp(const char *opt, const char *short_opt, const char *long_opt)
+{
+    return strcmp(opt, short_opt) == 0 || strcmp(opt, long_opt) == 0;
+}
+
 int
 main(int argc, const char **argv)
 {
+    char *p;
     unsigned int i;
     int rc;
+    unsigned long network_prefix = 32;
+    unsigned int suffix;
     progname = strrchr(argv[0], '/');
     if (progname)
 	progname++;
@@ -507,20 +571,56 @@ main(int argc, const char **argv)
 	progname = argv[0];
     openlog(progname, LOG_PID, LOG_MAIL);
     for (i = 1; i < argc; i++) {
-	if (strcmp("-d", argv[i]) == 0)
+	if (optp(argv[i], "-d", "--debug"))
 	    debug_me = 1;
-	else if (strcmp("-v", argv[i]) == 0)
+	else if (optp(argv[i], "-v", "--verbose"))
 	    opt_verbose++;
-	else if (strcmp("-h", argv[i]) == 0) {
+	else if (optp(argv[i], "-b", "--bloc-max-idle")) {
 	    if (++i >= argc)
-		fatal("Missing argument to -h");
+		fatal("Missing argument to --bloc-max-idle");
+	    bloc_max_idle = strtoul(argv[i], &p, 10);
+	    if (*p)
+		fatal("Invalid argument to --bloc-max-idle.  "
+		      "Integer value expected");
+	}
+	else if (optp(argv[i], "-g", "--greylist-delay")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --greylist-delay");
+	    greylist_delay = strtoul(argv[i], &p, 10);
+	    if (*p)
+		fatal("Invalid argument to --greylist-delay.  "
+		      "Integer value expected");
+	}
+	else if (optp(argv[i], "-h", "--home")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --home");
 	    db_home = argv[i];
 	}
+	else if (optp(argv[i], "-/", "--network-prefix")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --network-prefix");
+	    network_prefix = strtoul(argv[i], &p, 10);
+	    if (*p || network_prefix > 32U)
+		fatal("Invalid argument to --network-prefix.  "
+		      "Integer value between 0 and 32 expected");
+	}
+	else if (optp(argv[i], "-p", "--pass-max-idle")) {
+	    if (++i >= argc)
+		fatal("Missing argument to --pass-max-idle");
+	    pass_max_idle = strtoul(argv[i], &p, 10);
+	    if (*p)
+		fatal("Invalid argument to --pass-max-idle.  "
+		      "Integer value expected");
+	}
 	else {
 	    fprintf(stderr, "Unknown option \"%s\"\n", argv[i]);
 	    exit(EXIT_FAILURE);
 	}
     }
+    suffix = 32 - network_prefix;
+    ipv4_network_strip_bytes = suffix >> 3;
+    ipv4_network_mask = 0xffU & ~0U << (suffix & 0x7U);
+
     if (opt_verbose)
         syslog(LOG_NOTICE, "daemon started");
 #ifdef SIGHUP
